import { assertOptionalNumber, assertOptionalString } from '../validation.js';
import { bangUtils } from '../global-code/core.js';
import { coldFloatInletWithSetter } from '../standard-message-receivers.js';
import { computeUnitInSamples } from '../global-code/timing.js';
import { ast, ConstVar, AnonFunc, Var, Sequence, Class, Func } from '../../../../WebPd_compiler/dist/src/ast/declare.js';
import { commonsWaitEngineConfigure, commonsWaitFrame } from '../../../../WebPd_compiler/dist/src/stdlib/commons.js';
import '../../../../WebPd_compiler/dist/src/stdlib/fs.js';
import '../../../../WebPd_compiler/dist/src/compile/precompile/nodes.js';
import { generateVariableNamesNodeType } from '../variable-names.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
// ------------------------------- node builder ------------------------------ //
const builder = {
    translateArgs: (pdNode) => ({
        delay: assertOptionalNumber(pdNode.args[0]) || 0,
        unitAmount: assertOptionalNumber(pdNode.args[1]) || 1,
        unit: assertOptionalString(pdNode.args[2]) || 'msec',
    }),
    build: () => ({
        inlets: {
            '0': { type: 'message', id: '0' },
            '1': { type: 'message', id: '1' },
        },
        outlets: {
            '0': { type: 'message', id: '0' },
        },
        isPushingMessages: true
    }),
};
// ------------------------------- node implementation ------------------------------ //
const variableNames = generateVariableNamesNodeType('delay', [
    'setDelay',
    'scheduleDelay',
    'stop',
]);
const nodeImplementation = {
    initialization: ({ node: { args }, state, globs }) => ast `
        ${ConstVar(variableNames.stateClass, state, `{
            delay: 0,
            sampleRatio: 1,
            scheduledBang: SKED_ID_NULL,
        }`)}

        commons_waitEngineConfigure(() => {
            ${state}.sampleRatio = computeUnitInSamples(${globs.sampleRate}, ${args.unitAmount}, "${args.unit}")
            ${variableNames.setDelay}(${state}, ${args.delay})
        })
    `,
    messageReceivers: ({ state, globs, snds }) => ({
        '0': AnonFunc([Var('Message', 'm')]) `
            if (msg_getLength(m) === 1) {
                if (msg_isStringToken(m, 0)) {
                    ${ConstVar('string', 'action', 'msg_readStringToken(m, 0)')}
                    if (action === 'bang' || action === 'start') {
                        ${variableNames.scheduleDelay}(
                            ${state}, 
                            () => ${snds.$0}(msg_bang()),
                            ${globs.frame},
                        )
                        return
                    } else if (action === 'stop') {
                        ${variableNames.stop}(${state})
                        return
                    }
                    
                } else if (msg_isFloatToken(m, 0)) {
                    ${variableNames.setDelay}(${state}, msg_readFloatToken(m, 0))
                    ${variableNames.scheduleDelay}(
                        ${state},
                        () => ${snds.$0}(msg_bang()),
                        ${globs.frame},
                    )
                    return 
                }
            
            } else if (
                msg_isMatching(m, [MSG_STRING_TOKEN, MSG_FLOAT_TOKEN, MSG_STRING_TOKEN])
                && msg_readStringToken(m, 0) === 'tempo'
            ) {
                ${state}.sampleRatio = computeUnitInSamples(
                    ${globs.sampleRate}, 
                    msg_readFloatToken(m, 1), 
                    msg_readStringToken(m, 2)
                )
                return
            }
        `,
        '1': coldFloatInletWithSetter(variableNames.setDelay, state)
    }),
    dependencies: [
        computeUnitInSamples,
        bangUtils,
        commonsWaitEngineConfigure,
        commonsWaitFrame,
        () => Sequence([
            Class(variableNames.stateClass, [
                Var('Float', 'delay', '0'),
                Var('Float', 'sampleRatio', '1'),
                Var('SkedId', 'scheduledBang', 'SKED_ID_NULL'),
            ]),
            Func(variableNames.setDelay, [
                Var(variableNames.stateClass, 'state'),
                Var('Float', 'delay'),
            ], 'void') `
                state.delay = Math.max(0, delay)
            `,
            Func(variableNames.scheduleDelay, [
                Var(variableNames.stateClass, 'state'),
                Var('SkedCallback', 'callback'),
                Var('Int', 'currentFrame'),
            ], 'void') `
                if (state.scheduledBang !== SKED_ID_NULL) {
                    ${variableNames.stop}(state)
                }
                state.scheduledBang = commons_waitFrame(toInt(
                    Math.round(
                        toFloat(currentFrame) + state.delay * state.sampleRatio)),
                    callback
                )
            `,
            Func(variableNames.stop, [
                Var(variableNames.stateClass, 'state'),
            ], 'void') `
                commons_cancelWaitFrame(state.scheduledBang)
                state.scheduledBang = SKED_ID_NULL
            `
        ])
    ],
};

export { builder, nodeImplementation };
