import { assertOptionalString } from '../validation.js';
import { signalBuses } from '../global-code/buses.js';
import { ast, AnonFunc, Var, Sequence, Class, Func, ConstVar } from '../../../../WebPd_compiler/dist/src/ast/declare.js';
import '../../../../WebPd_compiler/dist/src/stdlib/fs.js';
import '../../../../WebPd_compiler/dist/src/compile/precompile/nodes.js';
import { generateVariableNamesNodeType } from '../variable-names.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
// ------------------------------- node builder ------------------------------ //
const builderThrow = {
    translateArgs: (pdNode) => ({
        busName: assertOptionalString(pdNode.args[0]) || '',
    }),
    build: () => ({
        inlets: {
            '0': { type: 'signal', id: '0' },
            '0_message': { type: 'message', id: '0_message' },
        },
        outlets: {},
        isPullingSignal: true,
    }),
    configureMessageToSignalConnection: (inletId) => {
        if (inletId === '0') {
            return { reroutedMessageInletId: '0_message' };
        }
        return undefined;
    },
};
const builderReceive = {
    translateArgs: (pdNode) => ({
        busName: assertOptionalString(pdNode.args[0]) || '',
    }),
    build: () => ({
        inlets: {
            '0': { type: 'message', id: '0' },
        },
        outlets: {
            '0': { type: 'signal', id: '0' },
        },
    }),
};
const builderSend = {
    translateArgs: (pdNode) => ({
        busName: assertOptionalString(pdNode.args[0]) || '',
    }),
    build: () => ({
        inlets: {
            '0': { type: 'signal', id: '0' },
        },
        outlets: {},
        isPullingSignal: true,
    }),
};
const builderCatch = {
    translateArgs: (pdNode) => ({
        busName: assertOptionalString(pdNode.args[0]) || '',
    }),
    build: () => ({
        inlets: {},
        outlets: {
            '0': { type: 'signal', id: '0' },
        },
    }),
};
// ------------------------------- generateDeclarations ------------------------------ //
const variableNames = generateVariableNamesNodeType('throw_catch_send_receive_t', ['setBusName']);
const nodeCore = () => Sequence([
    Class(variableNames.stateClass, [
        Var('string', 'busName'),
    ]),
    Func(variableNames.setBusName, [
        Var(variableNames.stateClass, 'state'),
        Var('string', 'busName')
    ], 'void') `
        if (busName.length) {
            state.busName = busName
            resetSignalBus(state.busName)
        }
    `
]);
const initialization = ({ node: { args }, state }) => ast `
        ${ConstVar(variableNames.stateClass, state, `{
            busName: '',
        }`)}

        ${variableNames.setBusName}(${state}, "${args.busName}")
    `;
// --------------------------------- throw~ ---------------------------------- //
const nodeImplementationThrow = {
    initialization: initialization,
    loop: ({ ins, state }) => ast `
        addAssignSignalBus(${state}.busName, ${ins.$0})
    `,
    messageReceivers: ({ state }) => ({
        '0_message': AnonFunc([Var('Message', 'm')], 'void') `
            if (
                msg_isMatching(m, [MSG_STRING_TOKEN, MSG_STRING_TOKEN])
                && msg_readStringToken(m, 0) === 'set'
            ) {
                ${variableNames.setBusName}(${state}, msg_readStringToken(m, 1))
                return
            }
        `
    }),
    dependencies: [
        signalBuses,
        nodeCore,
    ]
};
// --------------------------------- catch~ ---------------------------------- //
const nodeImplementationCatch = {
    initialization: initialization,
    loop: ({ outs, state, }) => ast `
        ${outs.$0} = readSignalBus(${state}.busName)
        resetSignalBus(${state}.busName)
    `,
    dependencies: [
        signalBuses,
        nodeCore
    ],
};
// --------------------------------- send~ ---------------------------------- //
const nodeImplementationSend = {
    initialization: initialization,
    loop: ({ state, ins }) => ast `
        setSignalBus(${state}.busName, ${ins.$0})
    `,
    dependencies: [
        signalBuses,
        nodeCore
    ],
};
// --------------------------------- receive~ ---------------------------------- //
const nodeImplementationReceive = {
    initialization: initialization,
    inlineLoop: ({ state }) => ast `readSignalBus(${state}.busName)`,
    messageReceivers: ({ state }) => ({
        '0': AnonFunc([Var('Message', 'm')]) `
            if (
                msg_isMatching(m, [MSG_STRING_TOKEN, MSG_STRING_TOKEN])
                && msg_readStringToken(m, 0) === 'set'
            ) {
                ${variableNames.setBusName}(${state}, msg_readStringToken(m, 1))
                return
            }
        `
    }),
    dependencies: [
        signalBuses,
        nodeCore,
    ]
};
// ------------------------------------------------------------------- //
const builders = {
    'throw~': builderThrow,
    'catch~': builderCatch,
    'send~': builderSend,
    'receive~': builderReceive,
};
const nodeImplementations = {
    'throw~': nodeImplementationThrow,
    'catch~': nodeImplementationCatch,
    'send~': nodeImplementationSend,
    'receive~': nodeImplementationReceive,
};

export { builders, nodeImplementations };
