import { mapArray, countTo } from '../../../../WebPd_compiler/dist/src/functional-helpers.js';
import { Sequence, Class, Var, Func, ConstVar, ast, AnonFunc } from '../../../../WebPd_compiler/dist/src/ast/declare.js';
import { fsWriteSoundStream } from '../../../../WebPd_compiler/dist/src/stdlib/fs.js';
import '../../../../WebPd_compiler/dist/src/compile/precompile/nodes.js';
import { assertOptionalNumber } from '../validation.js';
import { stringMsgUtils } from '../global-code/core.js';
import { parseSoundFileOpenOpts, parseReadWriteFsOpts } from '../global-code/fs.js';
import { generateVariableNamesNodeType } from '../variable-names.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
const BLOCK_SIZE = 44100 * 5;
// TODO: lots of things left to implement
// TODO : check the real state machine of writesf
//      - what happens when start / stopping / start stream ? 
//      - what happens when stream ended and starting again ? 
//      - etc ...
// ------------------------------- node builder ------------------------------ //
const builder = {
    translateArgs: (pdNode) => ({
        channelCount: assertOptionalNumber(pdNode.args[0]) || 1,
    }),
    build: ({ channelCount }) => ({
        inlets: {
            '0_message': { type: 'message', id: '0_message' },
            ...mapArray(countTo(channelCount), (channel) => [
                `${channel}`,
                { type: 'signal', id: `${channel}` },
            ]),
        },
        outlets: {},
        isPullingSignal: true,
    }),
    configureMessageToSignalConnection: (inletId) => {
        if (inletId === '0') {
            return { reroutedMessageInletId: '0_message' };
        }
        return undefined;
    },
};
// ------------------------------ generateDeclarations ------------------------------ //
const variableNames = generateVariableNamesNodeType('writesf_t', [
    'flushBlock'
]);
const nodeCore = () => Sequence([
    Class(variableNames.stateClass, [
        Var('fs_OperationId', 'operationId'),
        Var('boolean', 'isWriting'),
        Var('Array<FloatArray>', 'block'),
        Var('Int', 'cursor'),
    ]),
    Func(variableNames.flushBlock, [
        Var(variableNames.stateClass, 'state'),
    ], 'void') `
        ${ConstVar('Array<FloatArray>', 'block', '[]')}
        for (${Var('Int', 'i', '0')}; i < state.block.length; i++) {
            block.push(state.block[i].subarray(0, state.cursor))
        }
        fs_sendSoundStreamData(state.operationId, block)
        state.cursor = 0
    `,
]);
const initialization = ({ node: { args }, state }) => ast `
        ${ConstVar(variableNames.stateClass, state, `{
            operationId: -1,
            isWriting: false,
            cursor: 0,
            block: [
                ${countTo(args.channelCount).map(() => `createFloatArray(${BLOCK_SIZE})`).join(',')}
            ],
        }`)}
    `;
// ------------------------------- loop ------------------------------ //
const loop = ({ state, ins, node: { args } }) => ast `
    if (${state}.isWriting === true) {
        ${countTo(args.channelCount).map((i) => `${state}.block[${i}][${state}.cursor] = ${ins[i]}`)}
        ${state}.cursor++
        if (${state}.cursor === ${BLOCK_SIZE}) {
            ${variableNames.flushBlock}(${state})
        }
    }
`;
// ------------------------------- messageReceivers ------------------------------ //
const messageReceivers = ({ node, state, globs }) => ({
    '0_message': AnonFunc([Var('Message', 'm')], 'void') `
        if (msg_getLength(m) >= 2) {
            if (
                msg_isStringToken(m, 0) 
                && msg_readStringToken(m, 0) === 'open'
            ) {
                if (${state}.operationId !== -1) {
                    fs_closeSoundStream(${state}.operationId, FS_OPERATION_SUCCESS)
                }

                ${ConstVar('fs_SoundInfo', 'soundInfo', `{
                    channelCount: ${node.args.channelCount},
                    sampleRate: toInt(${globs.sampleRate}),
                    bitDepth: 32,
                    encodingFormat: '',
                    endianness: '',
                    extraOptions: '',
                }`)}
                ${ConstVar('Set<Int>', 'unhandledOptions', `parseSoundFileOpenOpts(
                    m,
                    soundInfo,
                )`)}
                ${ConstVar('string', 'url', `parseReadWriteFsOpts(
                    m,
                    soundInfo,
                    unhandledOptions
                )`)}
                if (url.length === 0) {
                    return
                }
                ${state}.operationId = fs_openSoundWriteStream(
                    url,
                    soundInfo,
                    () => {
                        ${variableNames.flushBlock}(${state})
                        ${state}.operationId = -1
                    }
                )
                return
            }

        } else if (msg_isAction(m, 'start')) {
                ${state}.isWriting = true
                return

        } else if (msg_isAction(m, 'stop')) {
            ${variableNames.flushBlock}(${state})
            ${state}.isWriting = false
            return

        } else if (msg_isAction(m, 'print')) {
            console.log('[writesf~] writing = ' + ${state}.isWriting.toString())
            return
        }    
    `
});
// ------------------------------------------------------------------- //
const nodeImplementation = {
    initialization: initialization,
    messageReceivers: messageReceivers,
    loop: loop,
    dependencies: [
        parseSoundFileOpenOpts,
        parseReadWriteFsOpts,
        stringMsgUtils,
        fsWriteSoundStream,
        nodeCore,
    ],
};

export { builder, nodeImplementation };
