import { Sequence, Var, Func, ast } from '../ast/declare.js';
import { toNodes } from '../dsp-graph/traversal.js';

const generateGlobs = ({ variableNamesIndex: { globs }, }) => 
// prettier-ignore
Sequence([
    Var('Int', globs.iterFrame, '0'),
    Var('Int', globs.frame, '0'),
    Var('Int', globs.blockSize, '0'),
    Var('Float', globs.sampleRate, '0'),
    Var('Float', globs.nullSignal, '0'),
    Func(globs.nullMessageReceiver, [
        Var('Message', 'm')
    ], 'void') ``
]);
/**
 * Embed arrays passed to the compiler in the compiled module.
 */
const generateEmbeddedArrays = ({ settings: { arrays } }) => Sequence(Object.entries(arrays).map(([arrayName, array]) => Sequence([
    `commons_setArray("${arrayName}", createFloatArray(${array.length}))`,
    `commons_getArray("${arrayName}").set(${JSON.stringify(Array.from(array))})`,
])));
const generateNodeInitializations = ({ precompilation, }) => Sequence([
    precompilation.traversals.all.map((nodeId) => precompilation.nodes[nodeId].initialization),
]);
const generateIoMessageReceivers = ({ settings: { io }, variableNamesIndex, }) => 
// Here not possible to assign directly the receiver because otherwise assemblyscript
// doesn't export a function but a global instead.
Sequence(Object.entries(io.messageReceivers).map(([nodeId, spec]) => spec.portletIds.map((inletId) => Func(variableNamesIndex.io.messageReceivers[nodeId][inletId], [Var('Message', 'm')], 'void') `${variableNamesIndex.nodes[nodeId].messageReceivers[inletId]}(m)`)));
const generateIoMessageSenders = ({ settings: { io }, variableNamesIndex }, generateIoMessageSender) => Sequence(Object.entries(io.messageSenders).map(([nodeId, spec]) => spec.portletIds.map((outletId) => {
    const listenerVariableName = variableNamesIndex.io.messageSenders[nodeId][outletId];
    return generateIoMessageSender(listenerVariableName, nodeId, outletId);
})));
const generatePortletsDeclarations = (compilation) => {
    const { graph, precompilation, settings: { debug }, } = compilation;
    const graphTraversalNodes = toNodes(graph, precompilation.traversals.all);
    return Sequence([
        graphTraversalNodes.map((node) => {
            const precompiledNode = precompilation.nodes[node.id];
            return [
                // 1. Declares signal outlets
                Object.values(precompiledNode.signalOuts).map((outName) => Var('Float', outName, '0')),
                // 2. Declares message receivers for all message inlets.
                Object.entries(precompiledNode.messageReceivers).map(([inletId, astFunc]) => {
                    // prettier-ignore
                    return Func(astFunc.name, astFunc.args, astFunc.returnType) `
                                ${astFunc.body}
                                throw new Error('[${node.type}], id "${node.id}", inlet "${inletId}", unsupported message : ' + msg_display(${astFunc.args[0].name})${debug
                        ? " + '\\nDEBUG : remember, you must return from message receiver'"
                        : ''})
                            `;
                }),
            ];
        }),
        // 3. Declares message senders for all message outlets.
        // This needs to come after all message receivers are declared since we reference them here.
        graphTraversalNodes.map((node) => Object.values(precompilation.nodes[node.id].messageSenders).map(({ messageSenderName: sndName, messageReceiverNames: rcvNames, }) => 
        // prettier-ignore
        Func(sndName, [
            Var('Message', 'm')
        ], 'void') `
                    ${rcvNames.map(rcvName => `${rcvName}(m)`)}
                `)),
    ]);
};
const generateLoop = (compilation) => {
    const { variableNamesIndex, precompilation } = compilation;
    const { traversals } = precompilation;
    const { globs } = variableNamesIndex;
    // prettier-ignore
    return ast `
        for (${globs.iterFrame} = 0; ${globs.iterFrame} < ${globs.blockSize}; ${globs.iterFrame}++) {
            _commons_emitFrame(${globs.frame})
            ${traversals.loop.map((nodeId) => precompilation.nodes[nodeId].loop)}
            ${globs.frame}++
        }
    `;
};
const generateImportsExports = ({ precompilation }, generateImport, generateExport) => Sequence([
    precompilation.dependencies.imports.map(generateImport),
    precompilation.dependencies.exports.map(generateExport),
]);

export { generateEmbeddedArrays, generateGlobs, generateImportsExports, generateIoMessageReceivers, generateIoMessageSenders, generateLoop, generateNodeInitializations, generatePortletsDeclarations };
