import compileToJavascript from '../engine-javascript/compile/index.js';
import compileToAssemblyscript from '../engine-assemblyscript/compile/index.js';
import { generateVariableNamesIndex } from './variable-names-index.js';
import { buildGraphTraversalAll } from './compile-helpers.js';
import { trimGraph } from '../dsp-graph/traversal.js';
import precompile, { initializePrecompilation } from './precompile/index.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
var index = (graph, nodeImplementations, target, compilationSettings) => {
    const settings = validateSettings(compilationSettings);
    const variableNamesIndex = generateVariableNamesIndex(graph, settings.debug);
    const graphTraversalAll = buildGraphTraversalAll(graph, settings.io);
    const trimmedGraph = trimGraph(graph, graphTraversalAll);
    const compilation = {
        graph: trimmedGraph,
        nodeImplementations,
        target,
        settings,
        variableNamesIndex,
        precompilation: initializePrecompilation(trimmedGraph, graphTraversalAll, variableNamesIndex),
    };
    precompile(compilation);
    let code;
    if (compilation.target === 'javascript') {
        code = compileToJavascript(compilation);
    }
    else if (compilation.target === 'assemblyscript') {
        code = compileToAssemblyscript(compilation);
    }
    else {
        throw new Error(`Invalid compilation.target ${compilation.target}`);
    }
    return {
        status: 0,
        code,
    };
};
/** Asserts user provided settings are valid (or throws error) and sets default values. */
const validateSettings = (compilationSettings) => {
    const arrays = compilationSettings.arrays || {};
    const io = {
        messageReceivers: (compilationSettings.io || {}).messageReceivers || {},
        messageSenders: (compilationSettings.io || {}).messageSenders || {},
    };
    const debug = compilationSettings.debug || false;
    const audio = compilationSettings.audio || {
        channelCount: { in: 2, out: 2 },
        bitDepth: 64,
    };
    if (![32, 64].includes(audio.bitDepth)) {
        throw new InvalidSettingsError(`"bitDepth" can be only 32 or 64`);
    }
    return {
        audio,
        arrays,
        io,
        debug,
    };
};
class InvalidSettingsError extends Error {
}

export { index as default, validateSettings };
