import { isGlobalDefinitionWithSettings, getNodeImplementation } from '../compile-helpers.js';
import { toNodes } from '../../dsp-graph/traversal.js';
import { Sequence } from '../../ast/declare.js';
import { commonsCore } from '../../stdlib/commons.js';
import { core } from '../../stdlib/core.js';
import '../../stdlib/fs.js';
import { msg } from '../../stdlib/msg.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
var precompileDependencies = (compilation) => {
    const { graph, nodeImplementations, precompilation, target, } = compilation;
    const dependencies = flattenDependencies([
        ...engineMinimalDependencies(),
        ..._collectDependenciesFromTraversal(nodeImplementations, graph, precompilation.traversals.all),
    ]);
    // Flatten and de-duplicate all the module's dependencies
    precompilation.dependencies.ast = instantiateAndDedupeDependencies(compilation, dependencies);
    // Collect and attach imports / exports info
    precompilation.dependencies.exports = collectAndDedupeExports(target, dependencies);
    precompilation.dependencies.imports = collectAndDedupeImports(dependencies);
};
const instantiateAndDedupeDependencies = (compilation, dependencies) => {
    const context = _getGlobalCodeGeneratorContext(compilation);
    return Sequence(dependencies
        .map((codeDefinition) => isGlobalDefinitionWithSettings(codeDefinition)
        ? codeDefinition.codeGenerator(context)
        : codeDefinition(context))
        .reduce((astElements, astElement) => astElements.every((otherElement) => !_deepEqual(otherElement, astElement))
        ? [...astElements, astElement]
        : astElements, []));
};
const engineMinimalDependencies = () => [
    core,
    commonsCore,
    msg,
];
const collectAndDedupeExports = (target, dependencies) => dependencies
    .filter(isGlobalDefinitionWithSettings)
    .filter((codeDefinition) => codeDefinition.exports)
    .reduce((exports, codeDefinition) => [
    ...exports,
    ...codeDefinition.exports.filter((xprt) => (!xprt.targets || xprt.targets.includes(target)) &&
        exports.every((otherExport) => xprt.name !== otherExport.name)),
], []);
const collectAndDedupeImports = (dependencies) => dependencies
    .filter(isGlobalDefinitionWithSettings)
    .filter((codeDefinition) => codeDefinition.imports)
    .reduce((imports, codeDefinition) => [
    ...imports,
    ...codeDefinition.imports.filter((imprt) => imports.every((otherImport) => imprt.name !== otherImport.name)),
], []);
const flattenDependencies = (dependencies) => dependencies.flatMap((codeDefinition) => {
    if (isGlobalDefinitionWithSettings(codeDefinition) &&
        codeDefinition.dependencies) {
        return [
            ...flattenDependencies(codeDefinition.dependencies),
            codeDefinition,
        ];
    }
    else {
        return [codeDefinition];
    }
});
const _collectDependenciesFromTraversal = (nodeImplementations, graph, graphTraversalAll) => {
    return toNodes(graph, graphTraversalAll)
        .reduce((definitions, node) => [
        ...definitions,
        ...getNodeImplementation(nodeImplementations, node.type)
            .dependencies,
    ], []);
};
const _getGlobalCodeGeneratorContext = ({ settings, target, variableNamesIndex }) => ({
    target: target,
    audioSettings: settings.audio,
    globs: variableNamesIndex.globs,
});
const _deepEqual = (ast1, ast2) => 
// This works but this flawed cause {a: 1, b: 2} and {b: 2, a: 1}
// would compare to false.
JSON.stringify(ast1) === JSON.stringify(ast2);

export { collectAndDedupeExports, collectAndDedupeImports, precompileDependencies as default, engineMinimalDependencies, flattenDependencies, instantiateAndDedupeDependencies };
